'use strict';

/**
 * Create a Game instance.
 */
function Game() {
	this.app = new Application({
		width: this.SCREEN_WIDTH,
		height: this.SCREEN_HEIGHT,
		antialias: true
	});
	
	this.scrollX = 0;
	this._scrollDirection = 0;
	
	/** {Vector2D} Acceleration due to gravity */
	this.gravity = new Vector2D(0, 0.2);
	
	// Define what game objects will exist.
	this._backgrounds = [];
	this._foreground;
	this._projectile;
	this._launcher;
	this._launchBar;
	this._enemies = [];
	
	// Create a promise to reflect the Pixi loader's state.
	this._resolveLoad;
	this._rejectLoad;
	this._loadPromise = new Promise((function (resolve, reject) {
		this._resolveLoad = resolve;
		this._rejectLoad = reject;
	}).bind(this));
	this._loadPromise.then((function () {
		delete(this._resolveLoad);
		delete(this._rejectLoad);
	}).bind(this));
	
	/** @private {Number} The amount of health the player has left */
	this._health = this.START_HEALTH;
	
	/** @private {Number} Counts down the time until the next enemy spawns, in milliseconds */
	this._enemySpawnTimer = 0;
	/** @private {Number} The number of enemies the player hit. */
	this._enemiesHit = 0;
	/** @private {Number} The level increases every 10 enemies */
	this._level = 0;
	
	/** @private {HUD} The manager for heads-up display elements */
	this._hud = new HUD(this);
	
	/** {Boolean} Whether sound is currently disabled */
	this.muted = false;
	
	/** @private {Number} The current state of the game, from `Game.prototype.STATES` */
	this._state = this.STATES.IDLE;
	/** @private {Number} The state of the game before it was paused, set on pause */
	this._prePauseState;
	
	this.app.view.tabIndex = 0;
	this.app.view.setAttribute('touch-action', 'none');
	this.app.view.addEventListener('pointermove',  this._handlePointerMove.bind(this), false);
	this.app.view.addEventListener('pointerleave', this._handlePointerLeave.bind(this), false);
	this.app.view.addEventListener('pointerdown',  this._handlePointerDown.bind(this), false);
	this.app.view.addEventListener('pointerup',    this._handlePointerUp.bind(this), false);
	
	this._lastFrameTime = 0;
	this._boundUpdate = this._update.bind(this);
}

// Define constants.
/** @constant {Number} The width of the visible game screen, in pixels */
Game.prototype.SCREEN_WIDTH = 1024;
/** @constant {Number} The height of the visible game screen, in pixels */
Game.prototype.SCREEN_HEIGHT = 432;
/** @constant {Number} The width of the entire game space, in pixels */
Game.prototype.GAME_WIDTH = 1670;
/** @constant {Number} How much of the screen from each edge should trigger scrolling */
Game.prototype.SCREEN_SCROLL_PART = 0.125;
/** @constant {Number} The speed at which the screen scrolls, in pixels per frame */
Game.prototype.SCREEN_SCROLL_SPEED = 3.5;
/** @constant {Number} The projectile's starting x-position */
Game.prototype.PROJECTILE_X = 100;
/** @constant {Number} The projectile's starting y-position */
Game.prototype.PROJECTILE_Y = Game.prototype.SCREEN_HEIGHT - 100;
/** @constant {Number} The launcher's x-position */
Game.prototype.LAUNCHER_X = 126;
/** @constant {Number} The launcher's y-position */
Game.prototype.LAUNCHER_Y = Game.prototype.SCREEN_HEIGHT - 94;
/** @constant {Number} The launch bar's x-position */
Game.prototype.LAUNCH_BAR_X = 94;
/** @constant {Number} The launch bar's y-position */
Game.prototype.LAUNCH_BAR_Y = Game.prototype.SCREEN_HEIGHT - 194;
/** @constant {Number} The x-coordinate at which enemies spawn */
Game.prototype.ENEMY_SPAWN_X = Game.prototype.GAME_WIDTH + Game.prototype.SCREEN_WIDTH + 5;
/** @constant {Number} The distance between spawned enemies, in pixels */
Game.prototype.ENEMY_SPAWN_SPACING = 256;
/** @constant {Number} The x-coordinate at which enemies take the castle and win */
Game.prototype.ENEMY_WIN_X = 332;
/** @constant {Number} The points awarded for hitting an enemy */
Game.prototype.POINTS_PER_HIT = 50;
/** @constant {Number} The maximum score that will fit in the score display */
Game.prototype.MAX_SCORE = 9999999;
/** @constant {Number} The number of enemies to hit to level up */
Game.prototype.LEVEL_UP_FREQUENCY = 5;
/** @constant {Number} The maximum game level */
Game.prototype.MAX_LEVEL = 999;
/** @constant {Number} The amount of health the player starts with */
Game.prototype.START_HEALTH = 3;

Game.prototype.STATES = {
	IDLE: 0,
	ANGLE: 1,
	POWER: 2,
	LAUNCHED: 3,
	PAUSED: 4,
	END: 5
};

Object.defineProperties(Game.prototype, {
	score: {
		get: function () {
			return Math.min(this.MAX_SCORE, (this.POINTS_PER_HIT * this._enemiesHit));
		}
	}
});

/**
 * Load necessary game assets and resolves `_loadPromise` when all are loaded.
 * @returns {Promise} Resolves when all assets are loaded
 */
Game.prototype.load = function () {
	Background.queueAssets();
	if (window.Foreground) {
		Foreground.queueAssets();
	}
	Projectile.queueAssets();
	if (window.ArmLauncher) {
		ArmLauncher.queueAssets();
	} else if (window.ShakeLauncher) {
		ShakeLauncher.queueAssets();
	}
	LaunchBar.queueAssets();
	Enemy.queueAssets();
	
	loader.onProgress.add(this._handleLoadProgress);
	loader.onComplete.add((function () {
		this._hud.loadPromise.then((function () {
			document.getElementById('loading-bar').remove();
			document.getElementById('game-screen').appendChild(this.app.view);
			document.getElementById('game-screen').appendChild(this._hud.topBar);
			document.getElementById('game-screen').appendChild(this._hud.levelUpMessage);
			
			this._resolveLoad();
		}).bind(this));
	}).bind(this));
	loader.load();
	
	return this._loadPromise;
};

/**
 * @private
 * Update the progress bar as assets are loaded.
 * @param {Loader} loader
 */
Game.prototype._handleLoadProgress = function (loader) {
	document.getElementById('loading-bar').value = loader.progress;
};

/**
 * Start the game (as long as all assets are loaded).
 * @returns {Promise} Resolves when all assets are loaded and the game has started
 */
Game.prototype.start = function () {
	return new Promise((function (resolve, reject) {
		this._loadPromise.then((function () {
			this._backgrounds = [
				new Background(this, 0.45, 'back'),
				new Background(this, 1, 'front')
			];
			this._projectile = new Projectile(this, this.PROJECTILE_X, this.PROJECTILE_Y);
			if (window.ArmLauncher) {
				this._launcher = new ArmLauncher(this, this.LAUNCHER_X, this.LAUNCHER_Y);
			} else if (window.ShakeLauncher) {
				this._launcher = new ShakeLauncher(this, this.LAUNCHER_X, this.LAUNCHER_Y);
			}
			this._launchBar = new LaunchBar(this, this.LAUNCH_BAR_X, this.LAUNCH_BAR_Y);
			
			if (window.Foreground) {
				this._foreground = new Foreground(this);
			}
			
			this._prespawnEnemies();
			this._enemySpawnTimer = this._calculateEnemySpawnInterval();
			
			this._lastFrameTime = 0;
			requestAnimationFrame(this._boundUpdate);
			resolve();
		}).bind(this));
	}).bind(this));
};

/**
 * Restart the game if it is not currently running.
 */
Game.prototype.restart = function () {
	// Do not restart if the game is running.
	if (this._state !== this.STATES.END) {
		return;
	}
	
	// Reset the score, level, and health.
	this._enemiesHit = 0;
	this._level = 0;
	this._health = this.START_HEALTH;
	this._hud.updateScore(this.score, true);
	this._hud.updateLevel(this._level + 1, true);
	this._hud.updateHealth(this._health, true);
	
	// Reset elements to their original positions.
	this.scrollX = 0;
	this._projectile.x = this.PROJECTILE_X;
	this._projectile.y = this.PROJECTILE_Y;
	this._launcher.x = this.LAUNCHER_X;
	this._launcher.y = this.LAUNCHER_Y;
	this._launchBar.x = this.LAUNCH_BAR_X;
	this._launchBar.y = this.LAUNCH_BAR_Y;
	
	// Reset enemies.
	this._enemies.forEach(function (enemy) {
		enemy.kill(true);
	}, this);
	this._prespawnEnemies();
	this._enemySpawnTimer = this._calculateEnemySpawnInterval();
	
	// Restart the game loop.
	this._state = this.STATES.IDLE;
	this._lastFrameTime = 0;
	requestAnimationFrame(this._boundUpdate);
};

/**
 * Pause the game and fire the event listener, if any.
 */
Game.prototype.pause = function () {
	if (this._state === this.STATES.PAUSED) {
		// Do not pause if already paused.
		return;
	}
	
	this._prePauseState = this._state;
	this._state = this.STATES.PAUSED;
	this._lastFrameTime = 0;
	
	this._hud.togglePause(true);
	
	if (typeof(this.onpause) === 'function') {
		this.onpause();
	}
};

/**
 * Resume the game and fire the event listener, if any.
 */
Game.prototype.resume = function () {
	if (this._state !== this.STATES.PAUSED) {
		// Do not resume if not paused.
		return;
	}
	
	this._state = this._prePauseState;
	
	requestAnimationFrame(this._boundUpdate);
	
	this._hud.togglePause(false);
	
	if (typeof(this.onresume) === 'function') {
		this.onresume();
	}
};

/**
 * @private
 * Scroll faster or advance the firing state when the pointer goes down.
 * @param {PointerEvent} ev
 */
Game.prototype._handlePointerDown = function (ev) {
	ev.preventDefault();
	ev.stopPropagation();
	
	// Shift focus to the canvas.
	this.app.view.focus();
	
	var pointerX = Utils.getCanvasX(ev.pageX, this.app.view.parentElement);
	
	// If the the pointer is at the side of the screen, scroll.
	if (pointerX < this.SCREEN_SCROLL_PART * this.SCREEN_WIDTH && this.scrollX > 0) {
		this._scrollDirection = -2;
	} else if (pointerX > (1 - this.SCREEN_SCROLL_PART) * this.SCREEN_WIDTH && this.scrollX < this.GAME_WIDTH - 1) {
		this._scrollDirection = 2;
	} else {
		// If the pointer is not at the side of the screen, advance the firing state.
		if (this._state === this.STATES.IDLE) {
			this.switchToPowerMode();
			this._handlePointerMove(ev);
		}
	}
};

/**
 * @private
 * Scroll slower when the pointer is up.
 * @param {PointerEvent} ev
 */
Game.prototype._handlePointerUp = function (ev) {
	ev.preventDefault();
	ev.stopPropagation();
	
	// Set the scroll direction to -1/0/1 as would happen without the pointer being down.
	this._handlePointerMove(ev);
	
	// If the player started aiming, launch the projectile.
	if (this._state === this.STATES.POWER) {
		this.launchProjectile();
	}
};

/**
 * @private
 * Set the screen to scroll depending on the pointer position when the pointer moves.
 * @param {PointerEvent} ev
 */
Game.prototype._handlePointerMove = function (ev) {
	ev.preventDefault();
	ev.stopPropagation();
	
	var pointerX = Utils.getCanvasX(ev.pageX, this.app.view.parentElement),
		pointerY = Utils.getCanvasY(ev.pageY, this.app.view.parentElement);
	
	if (pointerX < this.SCREEN_SCROLL_PART * this.SCREEN_WIDTH) {
		this._scrollDirection = -1;
	} else if (pointerX > (1 - this.SCREEN_SCROLL_PART) * this.SCREEN_WIDTH) {
		this._scrollDirection = 1;
	} else {
		this._scrollDirection = 0;
		
		if (this._state === this.STATES.IDLE || this._state === this.STATES.POWER) {
			// Aim at the cursor.
			var aimAngle = -Math.atan2(
				-(pointerY - (this._launchBar.y + this._launchBar.height)),
				(pointerX - this._launchBar.x));
			// Shift it a tiny bit since the launch bar starts slightly above middle.
			aimAngle -= LaunchBar.prototype.METER_MIN_ROTATION;
			// Update the launch bar arrow.
			this._launchBar.angle = aimAngle;
		}
	}
};

/**
 * @private
 * Stop scrolling the screen when the pointer exits the game.
 * @param {PointerEvent} ev
 */
Game.prototype._handlePointerLeave = function (ev) {
	this._scrollDirection = 0;
};

/**
 * @private
 * Spawn the first few enemies for the start of the game.
 */
Game.prototype._prespawnEnemies = function () {
	var NUMBER_TO_SPAWN = 6;
	for (var i = NUMBER_TO_SPAWN; i >= 0; i--) {
		var spawnX = this.ENEMY_SPAWN_X - (i * this.ENEMY_SPAWN_SPACING);
		this.spawnEnemy(spawnX);
	}
};

/**
 * @private
 * Calculate the time between enemy spawns to maintain consistent spacing as enemy speeds increase.
 * @returns {Number} The time to next enemy spawn, in milliseconds
 */
Game.prototype._calculateEnemySpawnInterval = function () {
	return (
		(this.ENEMY_SPAWN_SPACING + Utils.randInt(0, 256)) *
		(1 / (Enemy.prototype.SPEED + Enemy.prototype.LEVEL_SPEED_INCREASE * this._level)) *
		(1000 / 60));
};

/**
 * Spawn an enemy at the enemy spawn point, reusing existing enemy objects when possible.
 * @param {Number} [x] - The x-coordinate the enemy should spawn at, if not the default
 */
Game.prototype.spawnEnemy = function (x) {
	x = x || this.ENEMY_SPAWN_X;
	// Look for an enemy to reuse.
	var newEnemy = this._enemies.filter(function (enemy) { return enemy.isDead; })[0],
		levelEnemyVariant = this._level % Enemy.prototype.TYPES.length;
	
	if (newEnemy) {
		// If there is an enemy to reuse, reset it.
		newEnemy.x = x;
		newEnemy.revive(this._level, levelEnemyVariant);
	} else {
		// If there is not an enemy to reuse, create a new one.
		newEnemy = new Enemy(this, x, this._level, levelEnemyVariant);
		this._enemies.push(newEnemy);
	}
	if (this._foreground) {
		// Put the foreground back on top.
		this._foreground.bringToFront();
	}
};

/**
 * @private
 * Kill all active enemies that are within one screen width of the launcher.
 */
Game.prototype._killOnScreenEnemies = function () {
	this._enemies.forEach(function (enemy) {
		if (enemy.isDyingOrDead || enemy.x + enemy.width > this.SCREEN_WIDTH) {
			// Skip dead enemies or enemies more than a screen away from the launcher.
			return;
		}
		
		// Damage each enemy twice in case it has extra health.
		enemy.kill();
		enemy.kill();
	}, this);
};

/**
 * @private
 * Replace all active enemies with enemies of the current game level.
 */
Game.prototype._levelUpAllEnemies = function () {
	this._enemies.forEach(function (enemy) {
		if (enemy.isDyingOrDead) {
			// Skip dead enemies.
			return;
		}
		
		var enemyX = enemy.x,
			enemyY = enemy.y;
		
		enemy.kill(true);
		this.spawnEnemy(enemyX, enemyY);
	}, this);
};

/**
 * Switch to aiming the projectile launch angle.
 */
Game.prototype.switchToAngleMode = function () {
	this._launchBar.switchToAngleMode();
	this._state = this.STATES.ANGLE;
};

/**
 * Switch to adjusting the projectile launch power.
 */
Game.prototype.switchToPowerMode = function () {
	this._launchBar.switchToPowerMode();
	this._state = this.STATES.POWER;
};

/**
 * Launch the projectile at the set angle and power, and play the launcher animation.
 */
Game.prototype.launchProjectile = function () {
	// Stop the launch bar.
	this._launchBar.switchToIdleMode();
	// Launch the projectile.
	var launchVelocity = Vector2D.fromPolar(this._launchBar.power, this._launchBar.angle);
	this._projectile.launch(launchVelocity);
	// PLay the launcher's launch animation.
	this._launcher.launch();
	this._state = this.STATES.LAUNCHED;
};

/**
 * Return the projectile to the launcher and let the player ready another launch.
 */
Game.prototype.resetProjectile = function () {
	this._projectile.reset();
	this._launchBar.resetPower();
	this._state = this.STATES.IDLE;
};

/**
 * @private
 * Update game objects.  This should only be called through `requestAnimationFrame`.
 * @param {DOMHighResTimeStamp} timestamp - The timestamp of the frame, in milliseconds
 */
Game.prototype._update = function (timestamp) {
	// Do not continue to update if the game is paused or has ended.
	if (this._state === this.STATES.PAUSED || this._state === this.STATES.END) {
		return;
	}
	
	// Calculate time since last frame.
	if (!this._lastFrameTime) {
		this._lastFrameTime = timestamp;
	}
	var deltaTime = timestamp - this._lastFrameTime;
	this._lastFrameTime = timestamp;
	
	// Update screen scroll.
	this.scrollX += this._scrollDirection * this.SCREEN_SCROLL_SPEED;
	this.scrollX = Math.max(0, Math.min(this.GAME_WIDTH - this.SCREEN_WIDTH, this.scrollX));
	
	// Update all game objects.
	this._backgrounds.forEach(function (background) {
		background.update(deltaTime);
	});
	if (this._foreground) {
		this._foreground.update(deltaTime);
	}
	this._projectile.update(deltaTime);
	this._launcher.update(deltaTime);
	this._launchBar.update(deltaTime);
	this._enemies.forEach(function (enemy) {
		// Skip dead enemies.
		if (enemy.isDead) {
			return;
		}
		
		// Update the enemy.
		enemy.update(deltaTime);
		
		// Check whether the projectile is hitting the enemy.
		if (Utils.isCircCollidingRect(
				this._projectile.x, this._projectile.y, this._projectile.radius,
				enemy.hitbox.x, enemy.hitbox.y, enemy.hitbox.width, enemy.hitbox.height)) {
			enemy.kill();
			this.resetProjectile();
			
			if (enemy.isDyingOrDead) {
				// If the hit enemy died, increment the kill count.
				this._enemiesHit++;
				// Update the score display.
				this._hud.updateScore(this.score);
				
				if (this._enemiesHit % this.LEVEL_UP_FREQUENCY === 0) {
					// Increase the level if there have been enough hits.
					this._level = Math.min(this.MAX_LEVEL, this._level + 1);
					// Show level up message.
					this._hud.showLevelUpMessage();
					// Update the level counter.
					this._hud.updateLevel(this._level + 1);
					// Kill all enemies within one screen width of the launcher.
					this._killOnScreenEnemies();
					// Level up all active enemies.
					this._levelUpAllEnemies();
				}
			}
		}
		
		// Check whether the enemy has hit.
		if (enemy.hitbox.x < this.ENEMY_WIN_X) {
			// Decrement health.
			
			
			// Update the health display.
			this._hud.updateHealth(this._health - 1, false, true);
			this._health--;
			
			// Kill the enemy that hit.
			enemy.kill(true);
			
			if (this._health <= 0) {
				// If the player died, end the game.
				this._state = this.STATES.END;
				if (typeof(this.ongameover) === 'function') {
					this.ongameover(this.score);
				}
			}
		}
	}, this);
	
	// Spawn enemies.
	this._enemySpawnTimer -= deltaTime;
	if (this._enemySpawnTimer <= 0) {
		this.spawnEnemy();
		this._enemySpawnTimer = this._calculateEnemySpawnInterval();
	}
	
	// Reset the projectile if it falls through the ground.
	if (this._projectile.y > this.SCREEN_HEIGHT || this._projectile.x > this.GAME_WIDTH) {
		this.resetProjectile();
	}
	
	requestAnimationFrame(this._boundUpdate);
};
