'use strict';

/**
 * Create a new LaunchBar instance.
 * @param {Game} game - The game in which the LaunchBar is placed
 * @param {Number} x - The LaunchBar's starting x-coordinate
 * @param {Number} y - The LaunchBar's starting y-coordinate
 */
function LaunchBar(game, x, y) {
	this._game = game;
	
	/** @private {Sprite} The sprite for the arc the arrow moves along */
	this._arc = new Sprite(TextureCache[this.ARC_SPRITE])
	
	/** @private {Container} The wrapper for the arrow and its power meter fill */
	this._meterArrow = new Container();
	/** @private {Sprite} The sprite for the base arrow */
	this._meterOutline = new Sprite(TextureCache[this.METER_ARROW_SPRITE]);
	/** @private {Sprite} The sprite for the power meter fill */
	this._meterFill = new Sprite(TextureCache[this.METER_FILL_SPRITE]);
	
	// Add the arrow sprites to the wrapper.
	this._meterArrow.addChild(this._meterOutline);
	this._meterArrow.addChild(this._meterFill);
	
	// Set up how the arrow rotates and set it to its starting angle.
	this._meterArrow.pivot.x = this._meterFill.texture.width * this.METER_ROTATION_X;
	this._meterArrow.pivot.y = this._meterFill.texture.height * this.METER_ROTATION_Y;
	this._meterArrow.rotation = this.METER_MIN_ROTATION;
	
	/** @private {Number} The launch bar's x-coordinate in the world, not including screen scroll */
	this._baseX;
	
	// Move the launch bar to its position.
	this.x = x;
	this.y = y;
	
	// Set up the mask for the power meter fill.
	this._meterFillWidth = this._meterFill.texture.width;
	this._meterFillMask = new Rectangle(0, 0, 0, this._meterFill.texture.height);
	this._meterFill.texture.frame = this._meterFillMask;
	
	this._angleFrame = 0;
	this._powerFrame = 0;
	
	this._state = this.STATES.IDLE;
	
	game.app.stage.addChild(this._arc);
	game.app.stage.addChild(this._meterArrow);
}

// Define constants.
/** @constant {String} The path from the game root to the sprite for the aim arc */
LaunchBar.prototype.ARC_SPRITE = 'images/game/launch_arc.png';
/** @constant {String} The path from the game root to the sprite for the power meter arrow */
LaunchBar.prototype.METER_ARROW_SPRITE = 'images/game/launch_arrow.png';
/** @constant {String} The path from the game root to the sprite for the power meter fill */
LaunchBar.prototype.METER_FILL_SPRITE = 'images/game/launch_arrow_fill.png';
/** @constant {Number} The x-offset of the meter from the arc's x-position, in pixels */
LaunchBar.prototype.METER_X = -24;
/** @constant {Number} The y-offset of the meter from the arc's y-position, in pixels */
LaunchBar.prototype.METER_Y = 138;
/** @constant {Number} The x-position of the meter's rotation anchor point, from 0 to 1 */
LaunchBar.prototype.METER_ROTATION_X = -0.8;
/** @constant {Number} The y-position of the meter's rotation anchor point, from 0 to 1 */
LaunchBar.prototype.METER_ROTATION_Y = 0.5;
/** @constant {Number} The minimum rotation of the meter, in radians */
LaunchBar.prototype.METER_MIN_ROTATION = -0.01 * Math.TAU;
/** @constant {Number} The maximum rotation of the meter, in radians */
LaunchBar.prototype.METER_MAX_ROTATION = -0.175 * Math.TAU;
/** @constant {Number} The duration of the aim animation, in frames */
LaunchBar.prototype.ANGLE_DURATION = 120;
/** @constant {Number} The duration of the power animation, in frames */
LaunchBar.prototype.POWER_DURATION = 100;
/** @constant {Number} The minimum actual power value */
LaunchBar.prototype.POWER_MIN = 2;
/** @constant {Number} The maximum actual power value */
LaunchBar.prototype.POWER_MAX = 16;

LaunchBar.prototype.STATES = {
	IDLE: 0,
	ANGLE: 1,
	POWER: 2
};

/**
 * Queue assets to be preloaded that might be needed by any LaunchBar instance.
 */
LaunchBar.queueAssets = function () {
	loader.add(this.prototype.ARC_SPRITE);
	loader.add(this.prototype.METER_ARROW_SPRITE);
	loader.add(this.prototype.METER_FILL_SPRITE);
};

Object.defineProperties(LaunchBar.prototype, {
	x: {
		get: function () {
			return this._baseX;
		},
		set: function (value) {
			this._arc.x = value - this._game.scrollX;
			this._meterArrow.x = value + this.METER_X - this._game.scrollX;
			this._baseX = value;
		}
	},
	y: {
		get: function () {
			return this._arc.y;
		},
		set: function (value) {
			this._arc.y = value;
			this._meterArrow.y = value + this.METER_Y;
		}
	},
	width: {
		get: function () {
			return this._arc.width;
		}
	},
	height: {
		get: function () {
			return this._arc.height;
		}
	},
	angle: {
		get: function () {
			return this._meterArrow.rotation;
		},
		set: function (value) {
			this._meterArrow.rotation =
				Utils.constrainValue(value, this.METER_MAX_ROTATION, this.METER_MIN_ROTATION);
		}
	},
	power: {
		get: function () {
			return Utils.mappedAbsFunc(this._powerFrame, this.POWER_DURATION, this.POWER_MIN, this.POWER_MAX);
		}
	}
});

/**
 * Hide the power meter and switch the arrow to moving up and down the aim arc.
 */
LaunchBar.prototype.switchToAngleMode = function () {
	this.resetPower();
	this._state = this.STATES.ANGLE;
};

/**
 * Stop the arrow in place and switch to the power meter filling and emptying.
 */
LaunchBar.prototype.switchToPowerMode = function () {
	this.resetPower();
	this._state = this.STATES.POWER;
};

/**
 * Stop the arrow and power meter movement.
 */
LaunchBar.prototype.switchToIdleMode = function () {
	this._state = this.STATES.IDLE;
};

/**
 * CLear the power meter.
 */
LaunchBar.prototype.resetPower = function () {
	this._powerFrame = 0;
	this._meterFill.texture.frame.width = 0;
};

/**
 * Update the launch bar for the next frame.
 * @param {Number} deltaTime - The time since the last frame, in milliseconds
 */
LaunchBar.prototype.update = function (deltaTime) {
	// Update x in case the screen has scrolled.
	this.x = this.x;
	
	if (this._state === this.STATES.ANGLE) {
		this._angleFrame++;
		if (this._angleFrame >= this.ANGLE_DURATION) {
			this._angleFrame = 0;
		}
		this._meterArrow.rotation = Utils.mappedAbsFunc(this._angleFrame, this.ANGLE_DURATION, this.METER_MIN_ROTATION, this.METER_MAX_ROTATION);
	} else if (this._state === this.STATES.POWER) {
		this._powerFrame++;
		if (this._powerFrame >= this.POWER_DURATION) {
			this._powerFrame = 0;
		}
		this._meterFillMask.width = Utils.mappedAbsFunc(this._powerFrame, this.POWER_DURATION, 0, this._meterFillWidth);
		this._meterFill.texture.frame = this._meterFillMask;
	}
};
